from .base import ModelFlavor  # noqa
from .catboost import CatBoostModelFlavor  # noqa
from .custom import CustomModelFlavor  # noqa
from .huggingface import HuggingFaceModelFlavor  # noqa
from .keras import KerasModelFlavor  # noqa
from .lightgbm import LightGBMModelFlavor  # noqa
from .pytorch import PyTorchModelFlavor  # noqa
from .sklearn import ScikitLearnModelFlavor  # noqa
from .tensorflow import TensorFlowModelFlavor  # noqa
from .xgboost import XGBoostModelFlavor  # noqa
